//
//
//  Copyright 2013, Richard Gorton,     All Rights Reserved.
//
/// \file
/// Definitions for LWP context and ring buffer manipulation
///
/// \author Richard Gorton
///
/// \note:  For this particular implementation, have the constructor
///         allocate both the LWPCB and the Ring Buffer.  Another
///         implementation might choose a different policy.
///
///     The ring buffer can be located anywhere, but the spec recommends
///     the start of a page boundary.

#ifndef _LWPCONTEXT_H_
#define _LWPCONTEXT_H_

#include <limits>
#include <boost/noncopyable.hpp>

#include "lwpcore.h"

/// LWPContext - the combination of an LWPCB and a ring buffer for LWP Version 1
class LWPContext {
    public:

        /// Construct an LWP Context: the LWPCB and the ring buffer.
        /// \param[in]  number of ring buffer events (a minimum of 32) for LWP Version 1
        LWPContext(const size_t& numEvents)
          : m_ringBufferEvents(numEvents),
            m_ringBufferBytes(0),
            m_ringBuffer(NULL),
            m_LWPCB(NULL),
            m_EventConfigBase(NULL)
        {
            // Just in case it hasn't been done already
            LWPInit();

            m_ringBufferEvents = numEvents < (LWPMinBufferUnits() * 32) ? 32 : numEvents;
        }

        ~LWPContext()
        {
            if (NULL != m_LWPCB)
            {
                free(m_LWPCB);
            }
            m_LWPCB = NULL;
        }

        /// Create the ring buffer
        /// \returns    status code of the create.  In this implementation, one possibility is
        ///             LWPStatus_BadRingBufferAction (the producer thread must create the buffer)
        LWP_STATUS  CreateRingBuffer();

        /// Delete the ring buffer - should be performed by the harvester thread
        /// \returns    status code of the deletion.
        ///             LWPStatus_BadRingBufferAction - if the harvester thread was not the deleter
        LWP_STATUS  ReleaseRingBuffer();

        /// Perform all of allocations for the LWPContext.  Since the allocations
        /// might fail, put those in a separate function (best practices)
        /// This does NOT start or customize the LWPCB fields
        /// \returns    status code of the action
        LWP_STATUS  Initialize();

        /// Get the pointer to the LWPCB
        /// \returns    the pointer to the LWPCB
        LWPCB*      GetLWPCB()              { return m_LWPCB; }

        /// the size of the ring buffer (in numbers of events)
        /// \returns    the size of the ring buffer (in numbers of events)
        uint32_t    RingBufferEventSize()   { return m_ringBufferEvents; }

        /// the size of the ring buffer (in bytes)
        /// \returns    the size of the ring buffer (in bytes)
        uint32_t    RingBufferByteSize()    { return m_ringBufferBytes; }

    private:
        size_t              m_ringBufferEvents;
        size_t              m_ringBufferBytes;
        LWP_GenericEvent*   m_ringBuffer;   // Not quite true - the event records in the ring
                                            // buffer vary by type, but are all of the same size
        LWPCB*              m_LWPCB;
        LWPCB_Event*        m_EventConfigBase;  // First event configuration
};

#endif  // _LWPCONTEXT_H_
