//
//
//  Copyright 2012 - 2013, Richard Gorton,     All Rights Reserved.
//
/// \file
/// Definitions for the interface to the LWP (LightWeight Profiling) architectural
/// extension to the x86_64 architecture.  The LWP specification is desribed
/// in section 13.4 of the AMD64 Architecture Programmer’s Manual Volume 2: System Programming 
/// Caveat: AMD seems to move their web content around a lot, so you should google
/// One link that (presently) works is:
///     http://support.amd.com/us/Embedded_TechDocs/24593.pdf
/// This is a policy neutral (in terms of who manages the LWPCB and Ring Buffers)
/// implementation of a core set of features
///
/// \author Richard Gorton
///
/// \note   This implementation is 64-bit only.
///         Linux: not suppored by default in the kernel
///         Windows8: fully supported (including threshold mode) in the kernel
///         LWP REQUIRES the use of the XSAVE/XRSTOR instructions for context switching
///         See section 13.4.7 of the specification for details.
///
/// Hardware: The "bulldozer" cores support LWP.  An AMD FX-4170 processor implements:
///             LWPVAL, Instructions Retired, Branches Retired, Threshold
///

#ifndef _LWPCORE_H_
#define _LWPCORE_H_

#if defined(_WIN32)
#error  Needs development work to support windows (requires a Win8 development system)
/// Windows8 does have full OS support for LWP
#endif

#include <x86intrin.h>

#include <stdint.h>
#include <sys/types.h>

static const uint32_t   s_CPUID_XFeaturesSupported  = 0x0000000D;
static const uint32_t   s_CPUID_XFeaturesLWP        = 0x40000000;
static const uint32_t   s_CPUID_LWP                 = 0x8000001C;

/// LWP event ids.  The values correspond to Table 13-6 in the specification
typedef enum {
    LWPEventID_INVALID          = 0,
    LWPEventID_ValueSample      = 1,    /// Value Sample == LWPVAL instruction
    LWPEventID_InsnRetired      = 2,    /// Instructions retired
    LWPEventID_Branch_Retired   = 3,    /// Branches retired 
    LWPEventID_DCache_Miss      = 4,    /// DCache miss
    LWPEventID_CPU_Not_Halted   = 5,    /// CPU clocks NOT halted
    LWPEventID_CPU_Ref_Not_Halted = 6,  /// CPU reference clocks NOT halted (see the spec.)
    LWPEventID_Insert_User_Record = 255 /// Insert User Event Record in Ring Buffer == LWPINS
    } LWP_EVENTID;

/// Used for controlling event setup and threshold interrupts
typedef enum {
    LWPEventAction_INVALID      = 0,
    LWPEventAction_Enable,
    LWPEventAction_Disable
    } LWP_EVENTACTION;

/// used for continuous mode controls
typedef enum {
    LWPCMode_INVALID   = 0,
    LWPCMode_Continue,                  /// Use the continuous mode
    LWPCMode_Synchronized               /// Use synchronized mode
    } LWP_CMODE;

/// used for controlling the PTSC value
typedef enum {
    LWPPTSCMode_INVALID   = 0,
    LWPPTSCMode_Store,                  /// Use the continuous mode
    LWPPTSCMode_Zero                    /// Store zero
    } LWP_PTSCMODE;

/// Status of API interactions.
typedef enum {
    LWPStatus_INVALID           = 0,
    LWPStatus_Success,
    LWPStatus_EventOutOfRange,          /// specified event not in valid range for this machine
    LWPStatus_EventNotSupported,        /// specified event not supported on this system.
    LWPStatus_ContextRequiresStop,      /// The LWP context specified is active and must be
                                        /// stopped in order to perform this action
    LWPStatus_InvalidEventAction,       /// an invalid LWPCB_EVENTACTION was specified
    LWPStatus_EventIntervalOutOfRange,  /// event interval exceeds maximum permitted
    LWPStatus_EventCountOutOfRange,     /// event count exceeds maximum permitted
    LWPStatus_InvalidFilter,            /// Invalid filter value specified
    LWPStatus_FilterNotSupported,       /// The specified filter is not supported
    LWPStatus_RandomnessTooBig,         /// The randomization value is too large
    LWPStatus_FeatureNotAvailable,      /// The feature (continuous mode, PTSC support, threshold)
    LWPStatus_UnitializedVariable,      /// A parameter was unitialized
    LWPStatus_AllocationFailed,         /// Some sort of allocation of space failed
    LWPStatus_RingBufferTooBig,         /// A request was made to make the ring buffer bigger
                                        ///     larger than permitted
    LWPStatus_BadRingBufferAction,      /// A semantically improper action was done with
                                        /// respect to the ring buffer.  Implementation-specific
    LWPStatus_LWPNoHardware,            /// LWP is not supported by the hardware
    LWPStatus_LWPNotEnabledbyOS,        /// LWP is not supported and enabled by the OS
    } LWP_STATUS;


/// For a number of the various records, the descriptions in the documentation refer
/// to various bit widths.  Because of this, the sizes in the structures will be
/// defined as their uint<x>_t sizes.

/// Generic event record (Figure 13-20)
typedef struct {
    uint8_t     EventID;
    uint8_t     CoreID;
    uint16_t    Flags;          /// Event specific flags
    uint32_t    EventData1;
    uint64_t    EffectiveAddr;
    uint64_t    EventData2;
    uint64_t    TimeStamp;      /// Performance Time Stamp Counter
    } LWP_GenericEvent;

/// Programmed Value Sample - section 13.4.2.1
typedef struct {
    uint8_t     EventID;        /// EventID == LWPEventID_ValueSample
    uint8_t     CoreID;
    uint16_t    Flags;          /// Immediate value
    uint32_t    EventData1;     /// R/M value
    uint64_t    EffectiveAddr;
    uint64_t    EventData2;     /// Register value
    uint64_t    TimeStamp;
    } LWP_ValueSample;

/// Instructions Retired - section 13.4.2.2
typedef struct {
    uint8_t     EventID;        /// EventID == LWPEventID_InsnRetired
    uint8_t     CoreID;
    uint16_t    Reserved1;
    uint32_t    Reserved2;
    uint64_t    EffectiveAddr;  /// Instruction address
    uint64_t    Reserved3;
    uint64_t    TimeStamp;
    } LWP_InsnsRetired;

/// Branches Retires - section 13.4.2.3
typedef struct {
    uint8_t     EventID;        /// EventID == LWPEventID_Branch_Retired
    uint8_t     CoreID;
    uint16_t    Reserved1:12;   /// Reserved
    uint16_t    FusedOp:1;      /// Fused Instruction (FMA, etc): FUS in documentation
    uint16_t    PredValid:1;    /// Prediction Bit is valid: PRV in documentation
    uint16_t    Prediction:1;   /// Prediction Bit: PRD in documentation
    uint16_t    Taken:1;        /// Branch Taken Bit: TKN in documentation
    uint32_t    Reserved2;      /// Reserved
    uint64_t    EffectiveAddr;  /// Instruction address
    uint64_t    TargetAddress;  /// Target of instruction executed after branch
    uint64_t    TimeStamp;
    } LWP_BranchesRetired;

/// DCache Misses - section 13.4.2.4
typedef struct {
    uint8_t     EventID;        /// EventID == LWPEventID_DCache_Miss
    uint8_t     CoreID;
    uint16_t    Reserved:12;    /// Reserved
    uint16_t    DAddrValid:1;   /// Data Address is valid: DAV in documentation
    uint16_t    SourceLoc:3;    /// Source location for the data
    uint32_t    Latency;        /// Latency in cycles of the miss
    uint64_t    DataAddress;    /// Address of the reference which missed
    uint64_t    TimeStamp;
    } LWP_DCacheMisses;

/// CPU Clocks - a repeat of the InsnsRetired record
/// CPU Clocks not halted - section 13.4.2.5
/// The EventID == LWPEventID_CPU_Not_Halted
typedef LWP_InsnsRetired    LWP_CPUClocksNotHalted;

/// CPU Reference Clocks not halted - section 13.4.2.6
/// The EventID == LWPEventID_CPU_Ref_Not_Halted
typedef LWP_InsnsRetired    LWP_CPUReferenceClocksNotHalted;

/// [User] Programmed Event - the LWPINS instruction - section 13.4.2.7
typedef struct {
    uint8_t     EventID;        /// EventID == LWPEventID_Insert_User_Record
    uint8_t     CoreID;
    uint16_t    Flags;          /// Imm16 value
    uint32_t    Data1;          /// Reg/mem value
    uint64_t    EffectiveAddr;
    uint64_t    Data2;          /// Register value
    uint64_t    TimeStamp;
    } LWP_ProgrammedEvent;
/// synonym
typedef LWP_ProgrammedEvent LWP_UserEvent;

/// Future events - some ones mentioned are:
///     DTLB misses
///     FPU ops
///     ICache misses
///     ITLB misses


/// Detection of LWP via CPUID features (Table 13-7)
/// the LWP_FLAGS bits are part of the CPUID features, and mirrored in the LWPCB
typedef enum {
    LWP_FLAG_AVAILABLE          = 1 << 0,   /// LWP is supported both in hardware, and by the OS
                                            /// bit 62 of XFEATURE_ENABLED_MASK (XCR0 register)
    LWP_FLAG_VALUESAMPLE        = 1 << LWPEventID_ValueSample,
    LWP_FLAG_INSNRETIRED        = 1 << LWPEventID_InsnRetired,
    LWP_FLAG_BRANCHRETIRED      = 1 << LWPEventID_Branch_Retired,
    LWP_FLAG_DCACHE_MISS        = 1 << LWPEventID_DCache_Miss,
    LWP_FLAG_CPU_NOT_HALTED     = 1 << LWPEventID_CPU_Not_Halted,
    LWP_FLAG_CPU_REF_NOT_HALTED = 1 << LWPEventID_CPU_Ref_Not_Halted,
    /// Bits 28:7 are reserved
    /// Other mode bits not directly tied to EventIDs
    LWP_FLAG_CONTINUOUS         = 1 << 29,  /// Continuous sampling mode enabled
    LWP_FLAG_SAVE_PTSC          = 1 << 30,  /// Store performance timestamp counter
    LWP_FLAG_THRESHOLD_INT      = 1 << 31   /// Threshold interrupt mode
    } LWP_FLAGS;

/// what comes back from CPUID (EAX/EBX/ECX/EDX form)
typedef struct {
    /// EAX
    uint32_t    OSFeatures; /// LWP_FLAGS - the set of features which exist in hardware and are
                            ///     enabled by the OS.  See section 13.4.4.2.
                            ///     Note: if the features are not available in hardware, a #GP
                            ///     fault will occur on an attempt to set them as available.

    /// EBX
    uint8_t     LWPCB_Size;     /// Control block size (number of uint64_t)
    uint8_t     EventSize;      /// Byte size of an LWP Event record (32 bytes for V1)
    uint8_t     MaxEventID;     /// Max EventID value (excludes event 255).  Can be sparse
    uint8_t     EventOffset;    /// Offset from start of LWPCB to EventInterval1.

    /// ECX
    uint32_t    LatencyMax:5;   /// Number of bits in cache latency counters
    uint32_t    DataAddress:1;  /// Data Address is reported in DCacheMiss records
    uint32_t    LatencyRound:3; /// How much Cache latency is rounded
    uint32_t    Version:7;      /// Version of the implementation
    uint32_t    MinBufferSize:8;    /// Minimum permitted size of the event Ring Buffer
                                    /// in units of 32-event records
    uint32_t    Reserved:4;
        /// Filter control bits - can events be filtered on <type>
    uint32_t    FilterBranchPrediction:1;   /// BranchesRetired filtering
    uint32_t    FilterIP:1;                 /// IP filtering support
    uint32_t    FilterCacheLevels:1;        /// Cache event filtering based upon cache level
    uint32_t    FilterCacheLatency:1;       /// Cache event filtering based upon latency

    /// EDX
    uint32_t    HardwareFeatures;   /// LWP_FLAGS - what LWP features are supported by the processor
    } LWP_CPUID;

/// LWP_CFG configuration MSR - section 13.4.4.2
/// This is the means to enable/disable OS support for LWP, and is generally not
/// going to be something done by user-space code.  Listed here for completeness
typedef struct {
    uint32_t    EventsEnabled;  /// Corresponds to LWP_FLAGS
    uint8_t     CoreID;         /// What CoreID to write to event records (for virtualization)
    uint8_t     Vector;         /// Interrupt vector to use for LWP Threshold Interrupts
    uint16_t    Reserved;
    } LWP_CFGRegister;

/// Filter bits (used in the LWPCB) - Figure 13-31
static const uint32_t   s_AllFiltersMask = 0x7e001f00;
typedef enum {
    LWP_FILTER_CACHELEVEL               = 1 << 8,   /// Enable cache level filtering:
    LWP_FILTER_NORTHBRIDGE              = 1 << 9,   ///     Count Northbridge events
    LWP_FILTER_REMOTECACHE              = 1 << 10,  ///     Count remote events
    LWP_FILTER_RAM                      = 1 << 11,  ///     Count DRAM events
    LWP_FILTER_OTHER                    = 1 << 12,  ///     Count other events
    LWP_FILTER_NO_MISPREDICTEDBRANCHES  = 1 << 25,  /// No mispredicted branches
    LWP_FILTER_NO_PREDICTEDBRANCHES     = 1 << 26,  /// No predicted branches
    LWP_FILTER_NO_ABSBRANCHES           = 1 << 27,  /// No absolute branches
    LWP_FILTER_NO_CONDBRANCHES          = 1 << 28,  /// No conditional branches
    LWP_FILTER_NO_RELATIVEBRANCHES      = 1 << 29,  /// No relative branches
    LWP_FILTER_IP_INVERT                = 1 << 30,  /// Filter on instructions outside the IP range
    LWP_FILTER_IP_FILTER                = 1 << 31   /// Filtering on instruction inside the range
    } LWP_FILTERS;

/// LWPCB - LWP Control Block - section 13.4.6
typedef struct {
    uint32_t    Flags;          /// Corresponds to LWP_FLAGS
    uint32_t    BufferSize:28;  /// Size of the ringbuffer (bytes)
    uint32_t    Randomize:4;    /// Number of bits of randomness
    uint64_t    BufferBaseAddr; /// Base address of the ring buffer
    uint32_t    BufferHeadOffset;   /// Offset from base where the next event will be stored
    uint32_t    Reserved1;
    uint64_t    MissedEvents;   /// Count of missed events
    uint32_t    Threshold;      /// Threshold to signal an interrupt when ring buffer getting full
    uint32_t    Filters;        /// Event filter bits (corresponds to LWP_FILTERS)
    uint64_t    BaseIP;         ///     BaseIP for filtering
    uint64_t    LimitIP;        ///     LimitIP for filtering
    uint64_t    Reserved2;
    uint32_t    BufferTailOffset;   /// Offset to the oldest record in the ring buffer
    uint32_t    Reserved3;
    uint64_t    SWReserved[2];  /// Reserved for software use
    } LWPCB;
    /// After the last two reserved fields, there is some additional reserved space
    /// prior to the start of the event configuration area.  See EventOffset in CPUID:EBX

/// Event structures - section 13.4.6
static const int    s_EventValueLimit   = 0x01FFFFFF;
static const int    s_EventValueMaskVal = 0x03FFFFFF;
static const size_t s_MaxRingBufferSize = 0x0FFFFFFF;
typedef struct {
    uint32_t    EventInterval:26;   /// Reset value for counting events - treated as signed
    uint32_t    Reserved1:6;
    uint32_t    EventCounter:26;    /// Starting value of the event counter - treated as signed
    uint32_t    Reserved2:6;
    } LWPCB_Event;

extern LWP_CPUID    g_LWPCharacteristics;

/// Provide C language interface to the various routines
#ifdef __cplusplus
extern "C" {
#else
    #ifndef bool
    #define bool    int
    #endif

    #ifndef true
    #define true    (1 == 1)
    #endif

    #ifndef false
    #define false   (1 == 0)
    #endif
#endif

// ---------- Various Queries about LWP_CPUID content ----------

/// Is LWP available for use?
/// \returns    True if LWP is available in the hardware and enabled by the OS, false otherwise
static inline bool LWPAvailable()          { return (0 != (g_LWPCharacteristics.OSFeatures & LWP_FLAG_AVAILABLE)); }

/// Get the mask of all available features
/// \returns    the mask of all of the features available for use and supported by the OS
static inline const uint32_t LWPOSFeatures()   { return g_LWPCharacteristics.OSFeatures; }

/// Is a particular event available: LWPVAL, Instructions Retired, etc.
/// \returns    true if the event is available for use
bool LWPEventAvailable(LWP_EVENTID ev);

/// Is continuous mode sampling available
/// \returns    true if continuous mode is available for use
static inline bool LWPContinuousMode()     { return (0 != (LWP_FLAG_CONTINUOUS & g_LWPCharacteristics.OSFeatures)); }

/// Can the PTSC be saved in the record
/// \returns    true if the PTSC can be saved
static inline bool LWPSavePTSC()           { return (0 != (LWP_FLAG_SAVE_PTSC & g_LWPCharacteristics.OSFeatures)); }

/// Are threshold interrupts available
/// \returns    true if threshold interrupts available
static inline bool LWPThresholdInterrupts()       { return (0 != (LWP_FLAG_THRESHOLD_INT & g_LWPCharacteristics.OSFeatures)); }

/// Size (in uint64_t units) of the LWPCB.
/// \returns    the size (in uint64_t units) of the LWPCB
static inline const uint32_t LWPCBSizeInQuads()         { return g_LWPCharacteristics.LWPCB_Size; }

/// Size (in bytes) of the LWPCB.  It is stored as a number of uint64_t, but people
/// are accustomed to byte sizes generally
/// \returns    the size (in bytes) of the LWPCB
static inline uint32_t LWPCBSizeInBytes()         { return (sizeof(uint64_t) * g_LWPCharacteristics.LWPCB_Size); }

/// Size of an event record in the ring buffer.  (32 for V1 of LWP)
/// \returns    the size of event records
static inline const uint32_t LWPEventRecordSize()       { return g_LWPCharacteristics.EventSize; }

/// Maximum supported LWP_EVENTID (may be sparse) excluding LWPEventID_Insert_User_Record (LWPINS)
/// \returns    the maximum supported LWP_EVENTID
static inline const uint32_t LWPMaxEventID()            { return g_LWPCharacteristics.MaxEventID; }

/// Offset from the LWPCB to the EventInterval1 field
/// \returns    the offset to EventInterval1
static inline const uint32_t LWPEventOffset()           { return g_LWPCharacteristics.EventOffset; }

/// Number of bits in cache latency counters
/// \returns    the number of bits
static inline const uint32_t LWPLatencyMax()            { return g_LWPCharacteristics.LatencyMax; }

/// Do Cache miss enent record report the data address?
/// \returns    true if the cache miss event records report the miss address
static inline const bool LWPDataAddress()               { return g_LWPCharacteristics.DataAddress; }

/// Cache Latency rounding
/// \returns    the amount by which the cache latency is rounded
static inline const uint32_t LWPLatencyRound()          { return g_LWPCharacteristics.LatencyRound; }

/// Version of LWP implementation
/// \returns    the implementation version of LWP
static inline const uint32_t LWPVersion()               { return g_LWPCharacteristics.Version; }

/// Minimum size of the event LWP ring buffer (in units of 32 event records)
/// \returns    the raw number of buffer event units (not bytes)
static inline const uint32_t LWPMinBufferUnits()        { return g_LWPCharacteristics.MinBufferSize ; }

/// Minimum size of the event LWP ring buffer (in bytes)
/// \returns    the minimum size in bytes of the event buffer
static inline uint32_t LWPMinBufferSizeInBytes()  { return (32 * g_LWPCharacteristics.MinBufferSize * LWPEventRecordSize()); }

/// Can branches retired by filtered by correct prediction
/// \returns    true if branches retired events can be filtered based upon whether the branch was predicted properly
static inline const bool LWPBranchPredictionFiltering() { return g_LWPCharacteristics.FilterBranchPrediction; }

/// Is IP filtering supported
/// \returns    true if IP filtering is supported
static inline const bool LWPIPFiltering()               { return g_LWPCharacteristics.FilterIP; }

/// Can cache events be filtered by cache level
/// \returns    true if cache events can be filtered by cache level
static inline const bool LWPCacheLevelFiltering()       { return g_LWPCharacteristics.FilterCacheLevels; }

/// Can cache events be filtered by latency
/// \returns    true if cache events can be filtered by latency
static inline const bool LWPCacheLatencyFiltering()     { return g_LWPCharacteristics.FilterCacheLatency; }

/// Get the mask of available features supported by the processor
/// \returns    the mask of all of the features available for use and supported by the hardware
static inline const uint32_t LWPHardwareFeatures()    { return g_LWPCharacteristics.HardwareFeatures; }


// ---------- LWP Instruction wrappers ----------

/// Utility: LLWPCB instruction
/// \param[in]  pointer to Control block to load
static __inline__ void LoadLWPCB(LWPCB* cb)
{
#if 0
    __asm__("llwpcb %0"
            :
            : "r" (cb)  // input operand: cb
            );
#else
    // Just use the gcc intrinsic (remember to compile with "-mlwp")
    __llwpcb(cb);
#endif
}

/// Utility: SLWPCB instruction
/// \returns    pointer to Current LWPCB
static __inline__ LWPCB* CurrentLWPCB()
{
    LWPCB*  currCB;

#if 0
    __asm__("slwpcb %0"
            : "=r" (currCB) // output operand: currCB
            :
            );
#else
    // Use gcc intrinsic (remember to compile with "-mlwp")
    currCB = (LWPCB*)__slwpcb();
#endif

    return currCB;
}

/// Utility: LWPINS
/// This instruction is a bit tricky because the flags are physically encoded into the
/// instruction opcode as an immediate.  That is, callers must have flags be a literal
/// and not a variable reference.  Also see LWPVAL.
/// Special Note: the flags are encoded as uint32_t immediate in the instruction, but
/// the event record stores this by truncating to a uint16_t and storing it in the Flags
/// field.
/// \param[in]  flags
/// \param[in]  data1
/// \param[in]  data2
static __inline__ void EmitLWPINS(uint32_t flags, uint32_t data1, uint64_t data2)
{
    __lwpins64(data2, data1, flags);
}

/// Utility: LWPVAL
/// This instruction is a bit tricky because the flags are physically encoded into the
/// instruction opcode as an immediate.  That is, callers must have flags be a literal
/// and not a variable reference.  Also see LWPINS.
/// Special Note: the flags are encoded as uint32_t immediate in the instruction, but
/// the event record stores this by truncating to a uint16_t and storing it in the Flags
/// \param[in]  flags
/// \param[in]  data1
/// \param[in]  data2
static __inline__ void EmitLWPVAL(uint32_t flags, uint32_t data1, uint64_t data2)
{
    __lwpins64(data2, data1, flags);
}


// ---------- Queries about raw LWPCB values for a context ----------

/// Get the flags about which events are being counted (LWP_FLAGS)
/// \returns    the set of flags of the events being counted
static inline const uint32_t LWPContextFlags(LWPCB* cb)  { return cb->Flags; }

/// Get the total size of the event ring buffer for a context
/// \returns    size (in bytes) of the event ring buffer
static inline const uint32_t LWPContextBufferSize(LWPCB* cb) { return cb->BufferSize; }

/// Get the number of bits of randomization for a context
/// \returns    number of bits of randomization for a context
static inline const uint32_t LWPContextRandomize(LWPCB* cb)  { return cb->Randomize; }

/// Get the base address of the ring buffer
/// \returns    base address of the ring buffer
static inline const uint64_t LWPContextBufferBase(LWPCB* cb)    { return cb->BufferBaseAddr; }

/// Get the offset to the location where the next event will be stored
/// \returns    offset to the location where the next event will be stored
static inline const uint32_t LWPContextBufferHeadOffset(LWPCB* cb)  { return cb->BufferHeadOffset; }

/// Get the number of missed events
/// \returns    number of missed events
static inline const uint64_t LWPContextMissedEvents(LWPCB* cb)  { return cb->MissedEvents; }

/// Get the threshold at which to signal an interrupt as the buffer fills
/// \returns    threshold at which to signal an interrupt as the buffer fills
static inline const uint32_t LWPContextThreashHold(LWPCB* cb)   { return cb->Threshold; }

/// Get the event filter bits (corresponds to LWP_FILTERS)
/// \returns    event filter bits (corresponds to LWP_FILTERS)
static inline const uint32_t LWPContextFilters(LWPCB* cb)       { return cb->Filters; }

/// Get the BaseIP (address) for filtering
/// \returns    BaseIP (address) for filtering
static inline const uint64_t LWPContextBaseIP(LWPCB* cb)        { return cb->BaseIP; }

/// Get the LimitIP (address) for filtering
/// \returns    LimitIP (address) for filtering
static inline const uint64_t LWPContextLimitIP(LWPCB* cb)       { return cb->LimitIP; }

/// Get the offset to the location where the oldest record is located
/// \returns    offset to the location where the oldest record is located
static inline const uint32_t LWPContextBufferTailOffset(LWPCB* cb)  { return cb->BufferTailOffset; }

// ---------- Events and filters ----------

/// Get the event configuration for a type of event within the context; returning pointers to pointers
/// can be easily screwed up by consumer code, so avoid said pitfall.
/// \param[in]  pointer to the LWPCB
/// \param[in]  event
/// \param[out] pointer to status code
/// \returns    Pointer to the event itself
LWPCB_Event* LWPCBEvent(LWPCB* cb, const LWP_EVENTID ev, LWP_STATUS* theStatus);

/// Configure an event as enabled or disabled, set the interval/counter fields
/// \param[in]  pointer to the LWPCB
/// \param[in]  the enable/disable action
/// \param[in]  event interval
/// \param[in]  event counter
/// \returns    LWP_STATUS code
LWP_STATUS LWPConfigureEvent(LWPCB* cb, LWP_EVENTID ev, LWP_EVENTACTION act, uint32_t interval, uint32_t counter);

/// Enable an event
/// \param[in]  pointer to the LWPCB
/// \param[in]  event
/// \returns    status of the action
LWP_STATUS LWPEnableEvent(LWPCB* cb, LWP_EVENTID evid);

/// Disable an event
/// \param[in]  pointer to the LWPCB
/// \param[in]  event
/// \returns    status of the action
LWP_STATUS LWPDisableEvent(LWPCB* cb, LWP_EVENTID evid);

/// Set event filters
/// \param[in]  pointer to the LWPCB
/// \param[in]  filter value mask
/// \returns    status of the action
LWP_STATUS LWPSetFilters(LWPCB* cb, uint32_t filters);

/// Set minimum cache latency for events
/// \param[in]  pointer to the LWPCB
/// \param[in]  latency
/// \returns    status of the action
LWP_STATUS LWPSetLatency(LWPCB* cb, uint8_t filters);

// ---------- other operations to manipulate LWPCB context ----------
//
// The general model here is that the user must explicity manage all stops and [re]starts
// of LWP, rather than having it being done behind the scenes.
// 
/// Global setup  Must be called prior to usage of other features.  Need only be
/// called once.
/// \returns    True if LWP is available in the hardware and enabled by the OS, false otherwise
bool LWPInit();

/// Start LWP
/// \param[in]  pointer to a configured LWPCB to start
static inline void LWPStart(LWPCB* cb)    { LoadLWPCB(cb); }

/// Stop LWP
static inline void LWPStop()              { LoadLWPCB(NULL); }

/// Get the current LWP context
/// \returns    a pointer to the currently active LWPCB (NULL if none)
static inline LWPCB* LWPCurrentContext()  { return (CurrentLWPCB()); }

/// Is current LWP context active
/// \param[in]  pointer to an LWPCB
/// \returns    true if the specified context is active
static inline bool LWPContextIsActive(LWPCB* cb)
{
    LWPCB*  current;

    current = CurrentLWPCB();
    
    return (cb == current);
}

/// Set the ring buffer size
/// \param[in]  pointer to the LWPCB
/// \param[in]  size (in bytes)
/// \returns    status of the action
LWP_STATUS LWPSetRingBufferByteSize(LWPCB* cb, size_t nbytes);

/// Set the number of randomness bits (max is 15)
/// \param[in]  pointer to the LWPCB
/// \param[in]  amount to randomize
/// \returns    status of the action
LWP_STATUS LWPSetRandomness(LWPCB* cb, uint32_t rand);

/// Set the ring buffer size
/// \param[in]  pointer to the LWPCB
/// \param[in]  pointer to the ring buffer
/// \returns    status of the action
LWP_STATUS LWPSetRingBufferBaseAddress(LWPCB* cb, LWP_GenericEvent* buf);

/// Set the event threshold for interrupts
/// \param[in]  pointer to the LWPCB
/// \param[in]  number of bytes 
/// \returns    status of the action
LWP_STATUS LWPSetThreshold(LWPCB* cb, uint32_t nbytes);

/// Set the continuous mode
/// \param[in]  pointer to the LWPCB
/// \param[in]  the new mode
/// \returns    status of the action
LWP_STATUS LWPSetContinuousMode(LWPCB* cb, LWP_CMODE mode);

/// Set the PTSC mode
/// \param[in]  pointer to the LWPCB
/// \param[in]  the new mode
/// \returns    status of the action
LWP_STATUS LWPSetPTSCMode(LWPCB* cb, LWP_PTSCMODE mode);

/// Set the Interrupt threshold mode
/// \param[in]  pointer to the LWPCB
/// \param[in]  the new enabled/disabled setting
/// \returns    status of the action
LWP_STATUS LWPSetThresholdMode(LWPCB* cb, LWP_EVENTACTION action);

// ---------- Dynamic accesses of ring buffer ----------

/// Get the number of event records
/// \param[in]  pointer to the LWPCB
/// \returns    number of event records in the ring buffer
uint32_t LWPRecordsInRingBuffer(LWPCB* cb);

/// Set the ringbuffer tail offset (in terms of numbers of events)
/// \param[in]  pointer to the LWPCB
/// \param[in]  number of events currently in the ring
static inline void LWPSetRingBufferTailEvents(LWPCB* cb, uint32_t numEvents)
{
    cb->BufferTailOffset = numEvents * sizeof(LWP_GenericEvent);
}

#ifdef __cplusplus
}   // extern "C"
#endif

#endif  // _LWPCORE_H_
