//
//
//  Copyright 2013, Richard Gorton,     All Rights Reserved.
//
/// \file
/// Definitions for LWP context and ring buffer manipulation
///
/// \author Richard Gorton
///
/// \note:  For this particular implementation, have the constructor
///         allocate both the LWPCB and the Ring Buffer.  Another
///         implementation might choose a different policy.
///
///     The ring buffer can be located anywhere, but the spec recommends
///     the start of a page boundary.

#ifndef _IRETCONTEXT_H_
#define _IRETCONTEXT_H_

#include <limits>
#include <map>
#include <boost/noncopyable.hpp>

#include "lwpcore.h"
#include "context.h"

typedef void*       EventAddress_T;
typedef uint64_t    EventCount_T;
typedef std::map<EventAddress_T, EventCount_T> EventMap_T;

/// IRetContext - the derived context for a simple Instruction Return counter tool
class IRetContext : public LWPContext {
    public:
        /// Construct a Context for Counting Instruction Retired events
        /// \param[in]  number of ring buffer events (a minimum of 32) for LWP Version 1
        IRetContext(const size_t& numEvents)
          : LWPContext(numEvents),
            m_totalInsnRetEvents(0)
        {
            m_IRetEventMap.clear();
        }

        ~IRetContext()
        {
        }

        /// The number of instruction retired events processed
        uint64_t    InsnRetiredEvents()                     { return m_totalInsnRetEvents; }

        /// Increment the number of instruction retired events processed
        void        IncInsnRetiredEvents(uint64_t num)      { m_totalInsnRetEvents += num; }

        /// Return a reference to the IRetEventMap
        EventMap_T& IRetEventMap()                          { return m_IRetEventMap; }

    private:
        uint64_t            m_totalInsnRetEvents;   // Number of Instruction Retired events
        EventMap_T          m_IRetEventMap;
};

#endif  // _IRETCONTEXT_H_
